// Copyright (c) 2025-2026 Zensical and contributors

// SPDX-License-Identifier: MIT
// All contributions are certified under the DCO

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

// ----------------------------------------------------------------------------

//! Iterator over tokens conversions.

use std::borrow::Cow;

use super::Tokens;

// ----------------------------------------------------------------------------
// Traits
// ----------------------------------------------------------------------------

/// Borrow as [`Tokens`].
pub trait AsTokens {
    /// Borrows as an iterator over tokens.
    fn as_tokens(&self) -> Tokens<'_>;
}

// ----------------------------------------------------------------------------
// Trait implementations
// ----------------------------------------------------------------------------

impl AsTokens for &str {
    /// Borrows as an iterator over tokens.
    #[inline]
    fn as_tokens(&self) -> Tokens<'_> {
        Tokens::from(*self)
    }
}

impl AsTokens for String {
    /// Borrows as an iterator over tokens.
    #[inline]
    fn as_tokens(&self) -> Tokens<'_> {
        Tokens::from(self.as_str())
    }
}

impl AsTokens for Cow<'_, str> {
    /// Borrows as an iterator over tokens.
    #[inline]
    fn as_tokens(&self) -> Tokens<'_> {
        Tokens::from(self.as_ref())
    }
}
