// Copyright (c) 2025-2026 Zensical and contributors

// SPDX-License-Identifier: MIT
// All contributions are certified under the DCO

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the `Software`), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED `AS IS`, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

// ----------------------------------------------------------------------------

//! Segment.

use std::fmt::{self, Display};
use std::slice::Iter;

pub mod atom;
mod convert;
mod set;

pub use atom::Atom;
pub use convert::ToSegments;
pub use set::Segments;

// ----------------------------------------------------------------------------
// Structs
// ----------------------------------------------------------------------------

/// Segment.
#[derive(Clone, Debug, PartialEq, Eq)]
pub struct Segment<'a> {
    /// Atoms.
    atoms: Vec<Atom<'a>>,
}

// ----------------------------------------------------------------------------
// Implementations
// ----------------------------------------------------------------------------

impl Segment<'_> {
    /// Creates an iterator over the segment.
    #[inline]
    pub fn iter(&self) -> Iter<'_, Atom<'_>> {
        self.atoms.iter()
    }
}

// ----------------------------------------------------------------------------
// Trait implementations
// ----------------------------------------------------------------------------

impl<'a> FromIterator<Atom<'a>> for Segment<'a> {
    /// Creates a segment from an iterator.
    #[inline]
    fn from_iter<T>(iter: T) -> Self
    where
        T: IntoIterator<Item = Atom<'a>>,
    {
        Self {
            atoms: iter.into_iter().collect(),
        }
    }
}

impl<'a> IntoIterator for &'a Segment<'a> {
    type Item = &'a Atom<'a>;
    type IntoIter = Iter<'a, Atom<'a>>;

    /// Creates an iterator over the segment.
    #[inline]
    fn into_iter(self) -> Self::IntoIter {
        self.iter()
    }
}

// ----------------------------------------------------------------------------

impl Display for Segment<'_> {
    /// Formats the segment for display.
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        for atom in &self.atoms {
            Display::fmt(atom, f)?;
        }

        // No errors occurred
        Ok(())
    }
}
